/*
 * Copyright (c) 2001 Murray Stokely
 * Copyright (c) 2001 Wind River Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD$
 */

/*
 * This is a PCI shim for the Stallion EasyIO PCI Multi-port Serial Cards.
 * Stallion makes ISA and PCI versions of this card.
 */
#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/module.h>
#include <sys/systm.h>  /* uprintf */ 
#include <sys/errno.h>
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>  /* cdevsw, d_open_t, types */
#include <sys/malloc.h>
#include <sys/fcntl.h>		/* So we can get open modes */
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>	/* structs, prototypes for pci bus stuff */
#include <sys/rman.h>
#include <sys/tty.h>
#include <sys/ttydefaults.h>

#include <pci/pcivar.h> /* For get_pci macros! */
#include <pci/pcireg.h> /* For PCIR_ defines */

#include <dev/stl/comstats.h>
#include <dev/stl/stallion.h>
#include <dev/stl/stl_devices.h>

/* PCI Support Functions */

/*
 * Return identification string if this device is ours.
 */

static int
stl_probe(device_t dev)
{
	int err = ENXIO;
	uint16_t device_id;
	struct stl_device *pcicard;
	if (pci_get_vendor(dev) == STL_VENDOR_ID) {
		device_id = pci_get_device(dev);
		/* Iterate through the devices we know about to display the
		   relevant string.  (replaces 60 line switch statement) */
		for (pcicard = pci_boards; pcicard->device_id; pcicard++) {
			if (pcicard->device_id == device_id) {
				device_set_desc(dev, pcicard->description);
				err = 0;
				break;
			}
		}
	}

/*  Old driver didn't want to get confused with NatSemi chips? 

    class = pci_conf_read(tag, PCI_CLASS_REG);
    if ((class & PCI_CLASS_MASK) == PCI_CLASS_MASS_STORAGE)
	return((char *) NULL);
*/

	return err;
}


/* Attach function is only called if the probe is successful */

static int
stl_attach(device_t dev)
{
	int error = 0;
	struct stl_device *pcicard;
	uint16_t device_id;
	struct stl_softc *sc = device_get_softc(dev);

	sc->dev = dev;
	sc->unit = device_get_unit(dev);
	uprintf("stl_attach.  device_id : %d  unit : %d\n", 
	    pci_get_device(dev), sc->unit);

	/* Allocate the memory resource */
	sc->bar0id = 0x18;
	sc->bar0res = bus_alloc_resource(dev, SYS_RES_IOPORT, &(sc->bar0id),
	    0, ~0, 1, RF_ACTIVE);
	if (sc->bar0res == NULL) {
		uprintf("Memory allocation of PCI base register 0 failed!\n");
		error = ENXIO;
		goto fail1;
	}
	sc->bar1id = 0x14;
	sc->bar1res = bus_alloc_resource(dev, SYS_RES_IOPORT, &(sc->bar1id),
	    0, ~0, 1, RF_ACTIVE);
	if (sc->bar1res == NULL) {
		uprintf("Memory allocation of PCI base register 1 failed!\n");
		error =  ENXIO;
		goto fail2;
	}

	sc->bar0_bt = rman_get_bustag(sc->bar0res);
	sc->bar0_bh = rman_get_bushandle(sc->bar0res);
	sc->bar1_bt = rman_get_bustag(sc->bar1res);
	sc->bar1_bh = rman_get_bushandle(sc->bar1res);

/* We can't really do this until bus_Setup_intr has been called 
    sc->irq_bt = rman_get_bustag(sc->irqres);
    sc->irq_bh = rman_get_bushandle(sc->irqres);
*/


/* As of Sep 22, -STABLE still doesn't have rman_get_size */
#if __FreeBSD_version > 500000
/* It's really kind of useless, right?
	uprintf("BAR0 is %ld bytes\n", rman_get_size(sc->bar0res));
	uprintf("BAR1 is %ld bytes\n", rman_get_size(sc->bar1res));
*/
#endif

	/* Record the capabilities of this device */
	device_id = pci_get_device(dev);
	for (pcicard = pci_boards; pcicard->device_id; pcicard++) {
		if (pcicard->device_id == device_id) {
			sc->device_info = pcicard;
			break;
		}
	}

	/* Board initializtion, this finds out how many ports we have and
	   creates device nodes as well as sets up the proper interrupt
	   handler. */

	uprintf("Calling brdinit!\n");
	stl_brdinit(sc);

	/* Make the devices */
	stl_makedevs(sc);

	/* Allocate space for the data FIFO, this cannot return NULL */
/*    sc->fifo = (uint16_t *)malloc(ADC_FIFO_SIZE, M_NIFIFO, M_ZERO);*/

	/* Inititalize mutex */
/*    mtx_init(&sc->lock, device_get_nameunit(dev), MTX_DEF); */

	return error;

/*
 fail4:
    uprintf("fail4\n");
    bus_release_resource(dev, SYS_RES_IRQ, sc->irqid, sc->irqres);
 fail3:
    uprintf("fail3\n");
    bus_release_resource(dev, SYS_RES_MEMORY, sc->bar1id, sc->bar1res);
*/
fail2:
	uprintf("fail2\n");
	bus_release_resource(dev, SYS_RES_MEMORY, sc->bar0id, sc->bar0res);
fail1:
	uprintf("fail1\n");
	return error;
}




/* Detach device. */

static int
stl_detach(device_t dev)
{
	struct stl_softc *sc = device_get_softc(dev);
	int port = 0;
	uprintf("Releasing BAR0 resoruces!\n");
	/* Here we must release the memory resource */
	if (sc->bar0res) {
		bus_release_resource(dev, SYS_RES_IOPORT, sc->bar0id, sc->bar0res);
	}
	uprintf("Releasing BAR1 resoruces!\n");
	if (sc->bar1res) {
		bus_release_resource(dev, SYS_RES_IOPORT, sc->bar1id, sc->bar1res);
	}
	if (sc->irqres) { 
		/* XXX we never setup the intr */
		bus_teardown_intr(dev, sc->irqres, sc->handler); 
		bus_release_resource(dev, SYS_RES_IRQ, sc->irqid, sc->irqres);
	}

	/* Free the memory we allocated */
/*    free(sc->fifo, M_NIFIFO);
      free(sc->ai_channels, M_NICHAN); */
	uprintf("Destroying device nodes!\n");
	destroy_dev(sc->iomemdev);
	for (port = 0; port < sc->nrports; port++) {
		destroy_dev(sc->ttyE[port]);
		destroy_dev(sc->ttyiE[port]);
		destroy_dev(sc->ttylE[port]);
		destroy_dev(sc->cue[port]);
		destroy_dev(sc->cuie[port]);
		destroy_dev(sc->cule[port]);
	}
	return 0;
}

/* Called during system shutdown after sync. */

static int
stl_shutdown(device_t dev)
{
	uprintf("Stl shutdown!\n");
	return 0;
}


static device_method_t stl_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,	stl_probe),
	DEVMETHOD(device_attach, stl_attach),
	DEVMETHOD(device_detach, stl_detach),
	DEVMETHOD(device_shutdown, stl_shutdown),
	{ 0, 0 }
};

static driver_t stl_driver = {
	"stl",
	stl_methods,
	sizeof(struct stl_softc)
};

DRIVER_MODULE(stl, pci, stl_driver, stl_devclass, 0, 0);
