/*
 * stallion.c  -- stallion multiport serial driver.
 *
 * Copyright (c) 1995-1998 Greg Ungerer (gerg@stallion.oz.au).
 * Copyright (c) 2001 Wind River Systems.
 * Copyright (c) 2001 Murray Stokely <murray@FreeBSD.org>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Greg Ungerer.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/module.h>
#include <sys/systm.h>  /* uprintf */ 
#include <sys/errno.h>
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>  /* cdevsw, d_open_t, types */
#include <sys/malloc.h>
#include <sys/fcntl.h>		/* So we can get open modes */
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>	/* structs, prototypes for pci bus stuff */
#include <sys/rman.h>
#include <sys/tty.h>
#include <sys/ttydefaults.h>

#include <pci/pcivar.h> /* For get_pci macros! */
#include <pci/pcireg.h> /* For PCIR_ defines */

#include <dev/stl/comstats.h>
#include <dev/stl/stallion.h>
#include <dev/stl/stl_devices.h>

#include <assert.h>

/*****************************************************************************/
/*                      SC26198 HARDWARE FUNCTIONS                           */
/*****************************************************************************/

/*
 *	These functions get/set/update the registers of the sc26198 UARTs.
 *	Access to the sc26198 registers is via an address/data io port pair.
 *	(Maybe should make this inline...)
 */

int stl_sc26198getreg(stlport_t *portp, int regnr)
{
        struct stl_softc *sc = portp->sc;
	BAR0_WRITE1(sc, (portp->ioaddr + XP_ADDR), (regnr | portp->uartaddr));
	return(BAR0_READ1(sc, portp->ioaddr + XP_DATA));
}

void stl_sc26198setreg(stlport_t *portp, int regnr, int value)
{
	BAR0_WRITE1(portp->sc, (portp->ioaddr + XP_ADDR), (regnr | portp->uartaddr));
	BAR0_WRITE1(portp->sc, (portp->ioaddr + XP_DATA), value);
}

int stl_sc26198updatereg(stlport_t *portp, int regnr, int value)
{
        struct stl_softc *sc = portp->sc;
	BAR0_WRITE1(sc, (portp->ioaddr + XP_ADDR), (regnr | portp->uartaddr));
	if (BAR0_READ1(sc, portp->ioaddr + XP_DATA) != value) {
		BAR0_WRITE1(sc, (portp->ioaddr + XP_DATA), value);
		return(1);
	}
	return(0);
}

/*****************************************************************************/

/*
 *	Functions to get and set the sc26198 global registers.
 */

int stl_sc26198getglobreg(stlport_t *portp, int regnr)
{
        struct stl_softc *sc = portp->sc;
	BAR0_WRITE1(sc, (portp->ioaddr + XP_ADDR), regnr);
	return(BAR0_READ1(sc, portp->ioaddr + XP_DATA));
}

#if 0
void stl_sc26198setglobreg(stlport_t *portp, int regnr, int value)
{
        struct stl_softc *sc = portp->sc;
	BAR0_WRITE1(sc, (portp->ioaddr + XP_ADDR), regnr);
	BAR0_WRITE1(sc, (portp->ioaddr + XP_DATA), value);
}
#endif

/*****************************************************************************/

/*
 *	Inbitialize the UARTs in a panel. We don't care what sort of board
 *	these ports are on - since the port io registers are almost
 *	identical when dealing with ports.
 */

int stl_sc26198panelinit(struct stl_softc *sc, stlpanel_t *panelp)
{
	int	chipmask, i;
	int	nrchips, ioaddr;

#if DEBUG
	printf("stl_sc26198panelinit(brdp=%x,panelp=%x)\n", (int) brdp,
		(int) panelp);
#endif

	BRDENABLE(sc, panelp->pagenr);

/*
 *	Check that each chip is present and started up OK.
 */
	chipmask = 0;
	nrchips = (panelp->nrports + 4) / SC26198_PORTS;
	if (sc->brdtype == BRD_ECHPCI)
		BAR0_WRITE1(sc, 1, panelp->pagenr);

	for (i = 0; (i < nrchips); i++) {
		ioaddr = panelp->iobase + (i * 4); 
		BAR0_WRITE1(sc, (ioaddr + XP_ADDR), SCCR);
		BAR0_WRITE1(sc, (ioaddr + XP_DATA), CR_RESETALL);
		BAR0_WRITE1(sc, (ioaddr + XP_ADDR), TSTR);
		if (BAR0_READ1(sc, (ioaddr + XP_DATA)) != 0) {
			printf("STALLION: sc26198 not responding, "
				"board=%d panel=%d chip=%d\n", sc->unit,
				panelp->panelnr, i);
			continue;
		}
		chipmask |= (0x1 << i);
		BAR0_WRITE1(sc, (ioaddr + XP_ADDR), GCCR);
		BAR0_WRITE1(sc, (ioaddr + XP_DATA), GCCR_IVRTYPCHANACK);
		BAR0_WRITE1(sc, (ioaddr + XP_ADDR), WDTRCR);
		BAR0_WRITE1(sc, (ioaddr + XP_DATA), 0xff);
	}

	BRDDISABLE(sc);
	return(chipmask);
}

/*****************************************************************************/

/*
 *	Initialize hardware specific port registers.
 */

void stl_sc26198portinit(struct stl_softc *sc, stlpanel_t *panelp, stlport_t *portp)
{
#if DEBUG
	printf("stl_sc26198portinit(unit=%d,panelp=%x,portp=%x)\n",
		sc->unit, (int) panelp, (int) portp);
#endif

	if ((sc == (struct stl_softc *) NULL) || (panelp == (stlpanel_t *) NULL) ||
	    (portp == (stlport_t *) NULL))
		return;

	portp->ioaddr = panelp->iobase + ((portp->portnr < 8) ? 0 : 4);
	portp->uartaddr = (portp->portnr & 0x07) << 4;
	portp->pagenr = panelp->pagenr;
	portp->hwid = 0x1;

	BRDENABLE(sc, portp->pagenr);
	stl_sc26198setreg(portp, IOPCR, IOPCR_SETSIGS);
	BRDDISABLE(sc);
}

/*****************************************************************************/

/*
 *	Set up the sc26198 registers for a port based on the termios port
 *	settings.
 */

int stl_sc26198setport(stlport_t *portp, struct termios *tiosp)
{
	unsigned char	mr0, mr1, mr2, clk;
	unsigned char	imron, imroff, iopr, ipr;
	int		x;

#if DEBUG
	printf("stl_sc26198setport(portp=%x,tiosp=%x): brdnr=%d portnr=%d\n",
		(int) portp, (int) tiosp, portp->sc->unit, portp->portnr);
#endif

	mr0 = 0;
	mr1 = 0;
	mr2 = 0;
	clk = 0;
	iopr = 0;
	imron = 0;
	imroff = 0;

/*
 *	Set up the RX char ignore mask with those RX error types we
 *	can ignore.
 */
	portp->rxignoremsk = 0;
	if (tiosp->c_iflag & IGNPAR)
		portp->rxignoremsk |= (SR_RXPARITY | SR_RXFRAMING |
			SR_RXOVERRUN);
	if (tiosp->c_iflag & IGNBRK)
		portp->rxignoremsk |= SR_RXBREAK;

	portp->rxmarkmsk = SR_RXOVERRUN;
	if (tiosp->c_iflag & (INPCK | PARMRK))
		portp->rxmarkmsk |= (SR_RXPARITY | SR_RXFRAMING);
	if (tiosp->c_iflag & BRKINT)
		portp->rxmarkmsk |= SR_RXBREAK;

/*
 *	Go through the char size, parity and stop bits and set all the
 *	option registers appropriately.
 */
	switch (tiosp->c_cflag & CSIZE) {
	case CS5:
		mr1 |= MR1_CS5;
		break;
	case CS6:
		mr1 |= MR1_CS6;
		break;
	case CS7:
		mr1 |= MR1_CS7;
		break;
	default:
		mr1 |= MR1_CS8;
		break;
	}

	if (tiosp->c_cflag & CSTOPB)
		mr2 |= MR2_STOP2;
	else
		mr2 |= MR2_STOP1;

	if (tiosp->c_cflag & PARENB) {
		if (tiosp->c_cflag & PARODD)
			mr1 |= (MR1_PARENB | MR1_PARODD);
		else
			mr1 |= (MR1_PARENB | MR1_PAREVEN);
	} else {
		mr1 |= MR1_PARNONE;
	}

	mr1 |= MR1_ERRBLOCK;

/*
 *	Set the RX FIFO threshold at 8 chars. This gives a bit of breathing
 *	space for hardware flow control and the like. This should be set to
 *	VMIN.
 */
	mr2 |= MR2_RXFIFOHALF;

/*
 *	Calculate the baud rate timers. For now we will just assume that
 *	the input and output baud are the same. The sc26198 has a fixed
 *	baud rate table, so only discrete baud rates possible.
 */
	if (tiosp->c_ispeed == 0)
		tiosp->c_ispeed = tiosp->c_ospeed;
	if ((tiosp->c_ospeed < 0) || (tiosp->c_ospeed > SC26198_MAXBAUD))
		return(EINVAL);

	if (tiosp->c_ospeed > 0) {
		for (clk = 0; (clk < SC26198_NRBAUDS); clk++) {
			if (tiosp->c_ospeed <= sc26198_baudtable[clk])
				break;
		}
	}

/*
 *	Check what form of modem signaling is required and set it up.
 */
	if ((tiosp->c_cflag & CLOCAL) == 0) {
		iopr |= IOPR_DCDCOS;
		imron |= IR_IOPORT;
	}

/*
 *	Setup sc26198 enhanced modes if we can. In particular we want to
 *	handle as much of the flow control as possible automatically. As
 *	well as saving a few CPU cycles it will also greatly improve flow
 *	control reliability.
 */
	if (tiosp->c_iflag & IXON) {
		mr0 |= MR0_SWFTX | MR0_SWFT;
		imron |= IR_XONXOFF;
	} else {
		imroff |= IR_XONXOFF;
	}
#if 0
	if (tiosp->c_iflag & IXOFF)
		mr0 |= MR0_SWFRX;
#endif

	if (tiosp->c_cflag & CCTS_OFLOW)
		mr2 |= MR2_AUTOCTS;
	if (tiosp->c_cflag & CRTS_IFLOW)
		mr1 |= MR1_AUTORTS;

/*
 *	All sc26198 register values calculated so go through and set
 *	them all up.
 */

#if DEBUG
	printf("SETPORT: portnr=%d panelnr=%d brdnr=%d\n", portp->portnr,
		portp->panelnr, portp->sc->unit);
	printf("    mr0=%x mr1=%x mr2=%x clk=%x\n", mr0, mr1, mr2, clk);
	printf("    iopr=%x imron=%x imroff=%x\n", iopr, imron, imroff);
	printf("    schr1=%x schr2=%x schr3=%x schr4=%x\n",
		tiosp->c_cc[VSTART], tiosp->c_cc[VSTOP],
		tiosp->c_cc[VSTART], tiosp->c_cc[VSTOP]);
#endif

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	stl_sc26198setreg(portp, IMR, 0);
	stl_sc26198updatereg(portp, MR0, mr0);
	stl_sc26198updatereg(portp, MR1, mr1);
	stl_sc26198setreg(portp, SCCR, CR_RXERRBLOCK);
	stl_sc26198updatereg(portp, MR2, mr2);
	iopr = (stl_sc26198getreg(portp, IOPIOR) & ~IPR_CHANGEMASK) | iopr;
	if (tiosp->c_ospeed == 0) {
		iopr &= ~IPR_DTR;
	} else {
		iopr |= IPR_DTR;
		stl_sc26198setreg(portp, TXCSR, clk);
		stl_sc26198setreg(portp, RXCSR, clk);
	}
	stl_sc26198updatereg(portp, IOPIOR, iopr);
	stl_sc26198setreg(portp, XONCR, tiosp->c_cc[VSTART]);
	stl_sc26198setreg(portp, XOFFCR, tiosp->c_cc[VSTOP]);
	ipr = stl_sc26198getreg(portp, IPR);
	if (ipr & IPR_DCD)
		portp->sigs &= ~TIOCM_CD;
	else
		portp->sigs |= TIOCM_CD;
	portp->imr = (portp->imr & ~imroff) | imron;
	stl_sc26198setreg(portp, IMR, portp->imr);
	BRDDISABLE(portp->sc);
	portp->state &= ~(ASY_RTSFLOWMODE | ASY_CTSFLOWMODE);
	portp->state |= ((tiosp->c_cflag & CRTS_IFLOW) ? ASY_RTSFLOWMODE : 0);
	portp->state |= ((tiosp->c_cflag & CCTS_OFLOW) ? ASY_CTSFLOWMODE : 0);
	stl_ttyoptim(portp, tiosp);
	splx(x);

	return(0);
}

/*****************************************************************************/

/*
 *	Set the state of the DTR and RTS signals.
 */

void stl_sc26198setsignals(stlport_t *portp, int dtr, int rts)
{
	unsigned char	iopioron, iopioroff;
	int		x;

#if DEBUG
	printf("stl_sc26198setsignals(portp=%x,dtr=%d,rts=%d)\n",
		(int) portp, dtr, rts);
#endif

	iopioron = 0;
	iopioroff = 0;
	if (dtr == 0)
		iopioroff |= IPR_DTR;
	else if (dtr > 0)
		iopioron |= IPR_DTR;
	if (rts == 0)
		iopioroff |= IPR_RTS;
	else if (rts > 0)
		iopioron |= IPR_RTS;

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	if ((rts >= 0) && (portp->tty.t_cflag & CRTS_IFLOW)) {
		if (rts == 0) {
			stl_sc26198setreg(portp, MR1,
				(stl_sc26198getreg(portp, MR1) & ~MR1_AUTORTS));
			portp->stats.rxrtsoff++;
		} else {
			stl_sc26198setreg(portp, MR1,
				(stl_sc26198getreg(portp, MR1) | MR1_AUTORTS));
			portp->stats.rxrtson++;
		}
	}
	stl_sc26198setreg(portp, IOPIOR,
		((stl_sc26198getreg(portp, IOPIOR) & ~iopioroff) | iopioron));
	BRDDISABLE(portp->sc);
	splx(x);
}

/*****************************************************************************/

/*
 *	Return the state of the signals.
 */

int stl_sc26198getsignals(stlport_t *portp)
{
	unsigned char	ipr;
	int		x, sigs;

#if DEBUG
	printf("stl_sc26198getsignals(portp=%x)\n", (int) portp);
#endif

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	ipr = stl_sc26198getreg(portp, IPR);
	BRDDISABLE(portp->sc);
	splx(x);

	sigs = TIOCM_DSR;
	sigs |= (ipr & IPR_DCD) ? 0 : TIOCM_CD;
	sigs |= (ipr & IPR_CTS) ? 0 : TIOCM_CTS;
	sigs |= (ipr & IPR_DTR) ? 0: TIOCM_DTR;
	sigs |= (ipr & IPR_RTS) ? 0: TIOCM_RTS;
	return(sigs);
}

/*****************************************************************************/

/*
 *	Enable/Disable the Transmitter and/or Receiver.
 */

void stl_sc26198enablerxtx(stlport_t *portp, int rx, int tx)
{
	unsigned char	ccr;
	int		x;

#if DEBUG
	printf("stl_sc26198enablerxtx(portp=%x,rx=%d,tx=%d)\n",
		(int) portp, rx, tx);
#endif

	ccr = portp->crenable;
	if (tx == 0)
		ccr &= ~CR_TXENABLE;
	else if (tx > 0)
		ccr |= CR_TXENABLE;
	if (rx == 0)
		ccr &= ~CR_RXENABLE;
	else if (rx > 0)
		ccr |= CR_RXENABLE;

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	stl_sc26198setreg(portp, SCCR, ccr);
	BRDDISABLE(portp->sc);
	portp->crenable = ccr;
	splx(x);
}

/*****************************************************************************/

/*
 *	Start/stop the Transmitter and/or Receiver.
 */

void stl_sc26198startrxtx(stlport_t *portp, int rx, int tx)
{
	unsigned char	imr;
	int		x;

#if DEBUG
	printf("stl_sc26198startrxtx(portp=%x,rx=%d,tx=%d)\n",
		(int) portp, rx, tx);
#endif

	imr = portp->imr;
	if (tx == 0)
		imr &= ~IR_TXRDY;
	else if (tx == 1)
		imr |= IR_TXRDY;
	if (rx == 0)
		imr &= ~(IR_RXRDY | IR_RXBREAK | IR_RXWATCHDOG);
	else if (rx > 0)
		imr |= IR_RXRDY | IR_RXBREAK | IR_RXWATCHDOG;

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	stl_sc26198setreg(portp, IMR, imr);
	BRDDISABLE(portp->sc);
	portp->imr = imr;
	if (tx > 0) {
		portp->state |= ASY_TXBUSY;
		portp->tty.t_state |= TS_BUSY;
	}
	splx(x);
}

/*****************************************************************************/

/*
 *	Disable all interrupts from this port.
 */

void stl_sc26198disableintrs(stlport_t *portp)
{
	int	x;

#if DEBUG
	printf("stl_sc26198disableintrs(portp=%x)\n", (int) portp);
#endif

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	portp->imr = 0;
	stl_sc26198setreg(portp, IMR, 0);
	BRDDISABLE(portp->sc);
	splx(x);
}

/*****************************************************************************/

void stl_sc26198sendbreak(stlport_t *portp, long len)
{
	int	x;

#if DEBUG
	printf("stl_sc26198sendbreak(portp=%x,len=%d)\n",
		(int) portp, (int) len);
#endif

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	if (len == -1) {
		stl_sc26198setreg(portp, SCCR, CR_TXSTARTBREAK);
		portp->stats.txbreaks++;
	} else {
		stl_sc26198setreg(portp, SCCR, CR_TXSTOPBREAK);
	}
	BRDDISABLE(portp->sc);
	splx(x);
}

/*****************************************************************************/

/*
 *	Take flow control actions...
 */

void stl_sc26198sendflow(stlport_t *portp, int hw, int sw)
{
	unsigned char	mr0;
	int		x;

#if DEBUG
	printf("stl_sc26198sendflow(portp=%x,hw=%d,sw=%d)\n",
		(int) portp, hw, sw);
#endif

	if (portp == (stlport_t *) NULL)
		return;

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);

	if (sw >= 0) {
		mr0 = stl_sc26198getreg(portp, MR0);
		stl_sc26198setreg(portp, MR0, (mr0 & ~MR0_SWFRXTX));
		if (sw > 0) {
			stl_sc26198setreg(portp, SCCR, CR_TXSENDXOFF);
			mr0 &= ~MR0_SWFRX;
			portp->stats.rxxoff++;
		} else {
			stl_sc26198setreg(portp, SCCR, CR_TXSENDXON);
			mr0 |= MR0_SWFRX;
			portp->stats.rxxon++;
		}
		stl_sc26198wait(portp);
		stl_sc26198setreg(portp, MR0, mr0);
	}

	if (hw == 0) {
		portp->state |= ASY_RTSFLOW;
		stl_sc26198setreg(portp, MR1,
			(stl_sc26198getreg(portp, MR1) & ~MR1_AUTORTS));
		stl_sc26198setreg(portp, IOPIOR,
			(stl_sc26198getreg(portp, IOPIOR) & ~IOPR_RTS));
		portp->stats.rxrtsoff++;
	} else if (hw > 0) {
		portp->state &= ~ASY_RTSFLOW;
		stl_sc26198setreg(portp, MR1,
			(stl_sc26198getreg(portp, MR1) | MR1_AUTORTS));
		stl_sc26198setreg(portp, IOPIOR,
			(stl_sc26198getreg(portp, IOPIOR) | IOPR_RTS));
		portp->stats.rxrtson++;
	}

	BRDDISABLE(portp->sc);
	splx(x);
}

/*****************************************************************************/

/*
 *	Return the current state of data flow on this port. This is only
 *	really interresting when determining if data has fully completed
 *	transmission or not... The sc26198 interrupt scheme cannot
 *	determine when all data has actually drained, so we need to
 *	check the port statusy register to be sure.
 */

int stl_sc26198datastate(stlport_t *portp)
{
	unsigned char	sr;
	int		x;

#if DEBUG
	printf("stl_sc26198datastate(portp=%x)\n", (int) portp);
#endif

	if (portp == (stlport_t *) NULL)
		return(0);
	if (portp->state & ASY_TXBUSY) 
		return(1);

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	sr = stl_sc26198getreg(portp, SR);
	BRDDISABLE(portp->sc);
	splx(x);

	return((sr & SR_TXEMPTY) ? 0 : 1);
}

/*****************************************************************************/

void stl_sc26198flush(stlport_t *portp, int flag)
{
	int	x;

#if DEBUG
	printf("stl_sc26198flush(portp=%x,flag=%x)\n", (int) portp, flag);
#endif

	if (portp == (stlport_t *) NULL)
		return;

	x = spltty();
	BRDENABLE(portp->sc, portp->pagenr);
	if (flag & FWRITE) {
		stl_sc26198setreg(portp, SCCR, CR_TXRESET);
		stl_sc26198setreg(portp, SCCR, portp->crenable);
	}
	if (flag & FREAD) {
		while (stl_sc26198getreg(portp, SR) & SR_RXRDY)
			stl_sc26198getreg(portp, RXFIFO);
	}
	BRDDISABLE(portp->sc);
	splx(x);
}

/*****************************************************************************/

/*
 *	If we are TX flow controlled and in IXANY mode then we may
 *	need to unflow control here. We gotta do this because of the
 *	automatic flow control modes of the sc26198 - which downs't
 *	support any concept of an IXANY mode.
 */

void stl_sc26198txunflow(stlport_t *portp)
{
	unsigned char	mr0;

	mr0 = stl_sc26198getreg(portp, MR0);
	stl_sc26198setreg(portp, MR0, (mr0 & ~MR0_SWFRXTX));
	stl_sc26198setreg(portp, SCCR, CR_HOSTXON);
	stl_sc26198setreg(portp, MR0, mr0);
	portp->state &= ~ASY_TXFLOWED;
}

/*****************************************************************************/

/*
 *	Delay for a small amount of time, to give the sc26198 a chance
 *	to process a command...
 */

void stl_sc26198wait(stlport_t *portp)
{
	int	i;

#if DEBUG
	printf("stl_sc26198wait(portp=%x)\n", (int) portp);
#endif

	if (portp == (stlport_t *) NULL)
		return;

	for (i = 0; (i < 20); i++)
		stl_sc26198getglobreg(portp, TSTR);
}

/*****************************************************************************/

/*
 *	Transmit interrupt handler. This has gotta be fast!  Handling TX
 *	chars is pretty simple, stuff as many as possible from the TX buffer
 *	into the sc26198 FIFO.
 */

void stl_sc26198txisr(stlport_t *portp)
{
	unsigned int	ioaddr;
	unsigned char	mr0;
	char		*head, *tail;
	int		len, stlen;
	struct stl_softc *sc = portp->sc;

#if DEBUG
	printf("stl_sc26198txisr(portp=%x)\n", (int) portp);
#endif

	ioaddr = portp->ioaddr;

	head = portp->tx.head;
	tail = portp->tx.tail;
	len = (head >= tail) ? (head - tail) : (STL_TXBUFSIZE - (tail - head));
	if ((len == 0) || ((len < STL_TXBUFLOW) &&
	    ((portp->state & ASY_TXLOW) == 0))) {
		portp->state |= ASY_TXLOW;
		stl_dotimeout();
	}

	if (len == 0) {
		BAR0_WRITE1(sc, (ioaddr + XP_ADDR), (MR0 | portp->uartaddr));
		mr0 = BAR0_READ1(sc, ioaddr + XP_DATA);
		if ((mr0 & MR0_TXMASK) == MR0_TXEMPTY) {
			portp->imr &= ~IR_TXRDY;
			BAR0_WRITE1(sc, (ioaddr + XP_ADDR), (IMR | portp->uartaddr));
			BAR0_WRITE1(sc, (ioaddr + XP_DATA), portp->imr);
			portp->state |= ASY_TXEMPTY;
			portp->state &= ~ASY_TXBUSY;
		} else {
			mr0 |= ((mr0 & ~MR0_TXMASK) | MR0_TXEMPTY);
			BAR0_WRITE1(sc, (ioaddr + XP_DATA), mr0);
		}
	} else {
		len = MIN(len, SC26198_TXFIFOSIZE);
		portp->stats.txtotal += len;
		stlen = MIN(len, (portp->tx.endbuf - tail));
		BAR0_WRITE1(sc, (ioaddr + XP_ADDR), GTXFIFO);
		outsb((ioaddr + XP_DATA), tail, stlen);
		len -= stlen;
		tail += stlen;
		if (tail >= portp->tx.endbuf)
			tail = portp->tx.buf;
		if (len > 0) {
			outsb((ioaddr + XP_DATA), tail, len);
			tail += len;
		}
		portp->tx.tail = tail;
	}
}

/*****************************************************************************/

/*
 *	Receive character interrupt handler. Determine if we have good chars
 *	or bad chars and then process appropriately. Good chars are easy
 *	just shove the lot into the RX buffer and set all status byte to 0.
 *	If a bad RX char then process as required. This routine needs to be
 *	fast!
 */

void stl_sc26198rxisr(stlport_t *portp, unsigned int iack)
{
#if DEBUG
	printf("stl_sc26198rxisr(portp=%x,iack=%x)\n", (int) portp, iack);
#endif

	if ((iack & IVR_TYPEMASK) == IVR_RXDATA)
		stl_sc26198rxgoodchars(portp);
	else
		stl_sc26198rxbadchars(portp);

/*
 *	If we are TX flow controlled and in IXANY mode then we may need
 *	to unflow control here. We gotta do this because of the automatic
 *	flow control modes of the sc26198.
 */
	if ((portp->state & ASY_TXFLOWED) && (portp->tty.t_iflag & IXANY))
		stl_sc26198txunflow(portp);
}

/*****************************************************************************/

/*
 *	Process the good received characters from RX FIFO.
 */

void stl_sc26198rxgoodchars(stlport_t *portp)
{
	unsigned int	ioaddr, len, buflen, stlen;
	char		*head, *tail;
	struct stl_softc *sc = portp->sc;

#if DEBUG
	printf("stl_sc26198rxgoodchars(port=%x)\n", (int) portp);
#endif

	ioaddr = portp->ioaddr;

/*
 *	First up, calculate how much room there is in the RX ring queue.
 *	We also want to keep track of the longest possible copy length,
 *	this has to allow for the wrapping of the ring queue.
 */
	head = portp->rx.head;
	tail = portp->rx.tail;
	if (head >= tail) {
		buflen = STL_RXBUFSIZE - (head - tail) - 1;
		stlen = portp->rx.endbuf - head;
	} else {
		buflen = tail - head - 1;
		stlen = buflen;
	}

/*
 *	Check if the input buffer is near full. If so then we should take
 *	some flow control action... It is very easy to do hardware and
 *	software flow control from here since we have the port selected on
 *	the UART.
 */
	if (buflen <= (STL_RXBUFSIZE - STL_RXBUFHIGH)) {
		if (((portp->state & ASY_RTSFLOW) == 0) &&
		    (portp->state & ASY_RTSFLOWMODE)) {
			portp->state |= ASY_RTSFLOW;
			stl_sc26198setreg(portp, MR1,
				(stl_sc26198getreg(portp, MR1) & ~MR1_AUTORTS));
			stl_sc26198setreg(portp, IOPIOR,
				(stl_sc26198getreg(portp, IOPIOR) & ~IOPR_RTS));
			portp->stats.rxrtsoff++;
		}
	}

/*
 *	OK we are set, process good data... If the RX ring queue is full
 *	just chuck the chars - don't leave them in the UART.
 */
	BAR0_WRITE1(sc, (ioaddr + XP_ADDR), GIBCR);
	len = BAR0_READ1(sc, ioaddr + XP_DATA) + 1;
	if (buflen == 0) {
		BAR0_WRITE1(sc, (ioaddr + XP_ADDR), GRXFIFO);
/*XXX #assert "We haven't fixed this yet"	*/

/*		BAR0_READ2(sc, (ioaddr + XP_DATA), &stl_unwanted[0], len); */
		portp->stats.rxlost += len;
		portp->stats.rxtotal += len;
	} else {
		len = MIN(len, buflen);
		portp->stats.rxtotal += len;
		stlen = MIN(len, stlen);
		if (len > 0) {
			BAR0_WRITE1(sc, (ioaddr + XP_ADDR), GRXFIFO);
/* XXX "We haven't fixed this yet"		 */

/*			insb((ioaddr + XP_DATA), head, stlen); */
			head += stlen;
			if (head >= portp->rx.endbuf) {
				head = portp->rx.buf;
				len -= stlen;
/*XXX #assert "We haven't fixed this yet"	 */

/*				insb((ioaddr + XP_DATA), head, len); */
				head += len;
			}
		}
	}

	portp->rx.head = head;
	portp->state |= ASY_RXDATA;
	stl_dotimeout();
}



/*****************************************************************************/

/*
 *	Process all characters in the RX FIFO of the UART. Check all char
 *	status bytes as well, and process as required. We need to check
 *	all bytes in the FIFO, in case some more enter the FIFO while we
 *	are here. To get the exact character error type we need to switch
 *	into CHAR error mode (that is why we need to make sure we empty
 *	the FIFO).
 */

void stl_sc26198rxbadchars(stlport_t *portp)
{
	unsigned char	mr1;
	unsigned int	status;
	char		*head, *tail;
	char		ch;
	int		len;

/*
 *	First up, calculate how much room there is in the RX ring queue.
 *	We also want to keep track of the longest possible copy length,
 *	this has to allow for the wrapping of the ring queue.
 */
	head = portp->rx.head;
	tail = portp->rx.tail;
	len = (head >= tail) ? (STL_RXBUFSIZE - (head - tail) - 1) :
		(tail - head - 1);

/*
 *	To get the precise error type for each character we must switch
 *	back into CHAR error mode.
 */
	mr1 = stl_sc26198getreg(portp, MR1);
	stl_sc26198setreg(portp, MR1, (mr1 & ~MR1_ERRBLOCK));

	while ((status = stl_sc26198getreg(portp, SR)) & SR_RXRDY) {
		stl_sc26198setreg(portp, SCCR, CR_CLEARRXERR);
		ch = stl_sc26198getreg(portp, RXFIFO);

		if (status & SR_RXBREAK)
			portp->stats.rxbreaks++;
		if (status & SR_RXFRAMING)
			portp->stats.rxframing++;
		if (status & SR_RXPARITY)
			portp->stats.rxparity++;
		if (status & SR_RXOVERRUN)
			portp->stats.rxoverrun++;
		if ((portp->rxignoremsk & status) == 0) {
			if ((portp->tty.t_state & TS_CAN_BYPASS_L_RINT) &&
			    ((status & SR_RXFRAMING) ||
			    ((status & SR_RXPARITY) &&
			    (portp->tty.t_iflag & INPCK))))
				ch = 0;
			if ((portp->rxmarkmsk & status) == 0)
				status = 0;
			if (len > 0) {
				*(head + STL_RXBUFSIZE) = status;
				*head++ = ch;
				if (head >= portp->rx.endbuf)
					head = portp->rx.buf;
				len--;
			}
		}
	}

/*
 *	To get correct interrupt class we must switch back into BLOCK
 *	error mode.
 */
	stl_sc26198setreg(portp, MR1, mr1);

	portp->rx.head = head;
	portp->state |= ASY_RXDATA;
	stl_dotimeout();
}

/*****************************************************************************/

/*
 *	Other interrupt handler. This includes modem signals, flow
 *	control actions, etc.
 */

void stl_sc26198otherisr(stlport_t *portp, unsigned int iack)
{
	unsigned char	cir, ipr, xisr;
	int		rc;

#if DEBUG
	printf("stl_sc26198otherisr(portp=%x,iack=%x)\n", (int) portp, iack);
#endif

	cir = stl_sc26198getglobreg(portp, CIR);

	switch (cir & CIR_SUBTYPEMASK) {
	case CIR_SUBCOS:
		ipr = stl_sc26198getreg(portp, IPR);
		if (ipr & IPR_DCDCHANGE) {
			portp->state |= ASY_DCDCHANGE;
			portp->stats.modem++;
			stl_dotimeout();
		}
		break;
	case CIR_SUBXONXOFF:
		xisr = stl_sc26198getreg(portp, XISR);
		if (xisr & XISR_RXXONGOT) {
			portp->state |= ASY_TXFLOWED;
			portp->stats.txxoff++;
		}
		if (xisr & XISR_RXXOFFGOT) {
			portp->state &= ~ASY_TXFLOWED;
			portp->stats.txxon++;
		}
		break;
	case CIR_SUBBREAK:
		stl_sc26198setreg(portp, SCCR, CR_BREAKRESET);
		stl_sc26198rxbadchars(portp);
		break;
	default:
		break;
	}
}

/*****************************************************************************/

/*
 *	Interrupt service routine for sc26198 panels.
 */

void stl_sc26198intr(stlpanel_t *panelp, unsigned int iobase)
{
	stlport_t	*portp;
	unsigned int	iack;


/* 
 *	Work around bug in sc26198 chip... Cannot have A6 address
 *	line of UART high, else iack will be returned as 0.
 */

	return;

/* XXX 


	BAR0_WRITE1(sc, (iobase + 1), 0);

	iack = BAR0_READ1(sc, iobase + XP_IACK);
#if DEBUG
printf("stl_sc26198intr(panelp=%x,iobase=%x): iack=%x\n", panelp, iobase, iack);
#endif
	portp = panelp->ports[(iack & IVR_CHANMASK) + ((iobase & 0x4) << 1)];

	if (iack & IVR_RXDATA)
		stl_sc26198rxisr(portp, iack);
	else if (iack & IVR_TXDATA)
		stl_sc26198txisr(portp);
	else
		stl_sc26198otherisr(portp, iack);
*/
}

/*****************************************************************************/
